import os
import ray
import bz2
import tempfile
import ujson as json
from wrangl.data import FileDataset, Processor


@ray.remote
class MyProcessor(Processor):

    def process(self, raw):
        return json.loads(raw)


def create_file(num_lines, start=0):
    f = tempfile.NamedTemporaryFile(delete=False, suffix='.jsonl.bz2')
    f.close()
    data = [dict(value=start + i) for i in range(num_lines)]
    with bz2.open(f.name, 'wt') as fzip:
        for row in data:
            fzip.write(json.dumps(row) + '\n')
    return f.name, data


def load_files():
    fnames, data = [], []
    for i in range(3):
        fname_i, data_i = create_file(4, start=i*100)
        fnames.append(fname_i)
        data.extend(data_i)

    try:
        pool = ray.util.ActorPool([MyProcessor.remote() for _ in range(3)])
        loader = FileDataset(fnames, pool, cache_size=5)
        output = list(loader)
    finally:
        for f in fnames:
            os.remove(f)
    return data, output


if __name__ == '__main__':
    data, output = load_files()
    assert data == output
    print(output)



import os
import ray
import sqlite3
import tempfile
from wrangl.data import SQLDataset, Processor


@ray.remote
class MyProcessor(Processor):

    def process(self, raw):
        return dict(id=raw[0], value=raw[1])


def create_db(num_lines):
    f = tempfile.NamedTemporaryFile(delete=False, suffix='.db')
    f.close()
    data = [dict(value='a' * i, id=i) for i in range(num_lines)]
    db = sqlite3.connect(f.name, isolation_level=None)
    cursor = db.cursor()
    cursor.execute('CREATE TABLE data(id INTEGER PRIMARY KEY, value TEXT)')

    for row in data:
        cursor.execute('INSERT INTO data VALUES (?, ?)', (row['id'], row['value']))
    db.commit()
    cursor.close()
    db.close()
    return f.name, data


def load_db():
    fname, data = create_db(5)
    query = 'SELECT id, value FROM data ORDER BY id'

    try:
        pool = ray.util.ActorPool([MyProcessor.remote() for _ in range(3)])
        loader = SQLDataset(fname, query, pool, cache_size=5)
        output = list(loader)
    finally:
        os.remove(fname)
    return data, output


if __name__ == '__main__':
    data, output = load_db()
    assert data == output
    print(output)



import ray
import bz2
import sqlite3
from typing import Union, List
from collections.abc import Iterable
from pathlib import Path
from torch.utils.data import IterableDataset as Base

__all__ = ["Dataset", "IterableDataset", "SQLDataset", "FileDataset"]


class FullCacheException(Exception):
    pass


class Dataset(Base):
    """
    An interface that wraps `torch.utils.data.IterableDataset` to processes examples in parallel.
    """

    def __init__(self, pool: ray.util.actor_pool.ActorPool, cache_size: int = 1024, shuffle: bool = False, timeout: int = 10):
        """
        Args:
            pool: pool of `wrangl.data.loader.processor.Processor`s to process each examples.
            cache_size: how many examples to keep in the cache.
            shuffle: pseudo random order (local shuffling within cache).
            timeout: processor timeout.
        """
        self.pool = pool
        self.cache_size = cache_size
        self.shuffle = shuffle
        self.timeout = timeout
        self.in_pool = 0

    def add(self, o):
        """
        Adds an example `o` to be processed.

        Args:
            o: example to add.
        """
        if self.cache_is_full():
            raise FullCacheException('Cache is full at {}'.format(self.cache_size))
        self.pool.submit(lambda a, v: a.process.remote(v), o)
        self.in_pool += 1

    def cache_is_full(self):
        """
        Returns whether number of examples in the cache exceeds `self.cache_size`.
        """
        return self.in_pool >= self.cache_size

    def cache_is_empty(self):
        """
        Returns whether cache is empty.
        """
        return self.in_pool == 0

    def pop(self, ordered: bool = True, timeout: int = None):
        """
        Fetches a batch of processed examples from the cache.

        Args:
            ordered: whether to retrieve examples in insertion order.
            timeout: number of seconds to wait for retrieval. `None` means wait indefinitely.
        """
        get_next = self.pool.get_next if ordered else self.pool.get_next_unordered
        o = get_next(timeout=timeout)
        self.in_pool -= 1
        return o

    def iterate_unprocessed(self):
        """
        You must implement this function.
        Loads unprocessed examples.

        Returns:
            Iterator of the next example.
        """
        raise NotImplementedError()

    def __iter__(self):
        for o in self.iterate_unprocessed():
            if self.cache_is_full():
                yield self.pop(ordered=not self.shuffle, timeout=self.timeout)
            self.add(o)

        while not self.cache_is_empty():
            yield self.pop(ordered=not self.shuffle, timeout=self.timeout)

    def close(self):
        """
        Clean up any resources.
        """
        pass


class IterableDataset(Dataset):
    """
    A Dataset over a Python iterable.
    """

    def __init__(self, iterable: Iterable, pool: ray.util.ActorPool, cache_size: int = 1024, shuffle: bool = False, timeout: int = 10):
        """
        Processes list of examples.

        Args:
            iterable: list of examples to process.
            pool: pool of `wrangl.data.loader.processor.Processor`s to process each example. If not specified, then no processing will be done.
            cache_size: how many examples to keep in the cache.
            shuffle: pseudo random order (local shuffling within cache).
            timeout: processor timeout.
        """
        super().__init__(pool, cache_size=cache_size, shuffle=shuffle, timeout=timeout)
        self.iterable = iterable

    def iterate_unprocessed(self):
        """
        Loads unprocessed examples.

        Returns:
            Iterator of the next example.
        """
        for ex in self.iterable:
            yield ex


class SQLDataset(Dataset):
    """
    A Dataset over a SQLite SQL query.
    """

    def __init__(self, fdb: Union[str, Path], query: str, pool: ray.util.ActorPool, cache_size: int = 1024, shuffle: bool = False, timeout: int = 10):
        """
        Loads rows from SQL database.

        Args:
            fdb: path to SQL database.
            query: SQL query to retrieve data.
            pool: pool of `wrangl.data.loader.processor.Processor`s to process each example. If not specified, then no processing will be done.
            cache_size: how many examples to keep in the cache.
            shuffle: pseudo random order (local shuffling within cache).
            timeout: processor timeout.
        """
        super().__init__(pool, cache_size=cache_size, shuffle=shuffle, timeout=timeout)
        self.fdb = fdb
        self.query = query
        self.db = sqlite3.connect(self.fdb, isolation_level=None)
        self.cursor = self.db.cursor()
        self.rows = None

    def iterate_unprocessed(self):
        """
        Loads unprocessed examples.

        Returns:
            Iterator of the next example.
        """
        rows = self.cursor.execute(self.query)
        for row in rows:
            yield row

    def close(self):
        super().close()
        self.cursor.close()
        self.db.close()


class FileDataset(Dataset):
    """
    A Dataset over a list of files.
    """

    def __init__(self, fnames: List[Union[str, Path]], pool: ray.util.ActorPool, cache_size: int = 1024, shuffle: bool = False, timeout: int = 10):
        """
        Loads lines from input files.

        Args:
            fnames: list of text or `bz2` files to load from.
            pool: pool of `wrangl.data.loader.processor.Processor`s to process each example. If not specified, then no processing will be done.
            cache_size: how many examples to keep in the cache.
            shuffle: pseudo random order (local shuffling within cache).
            timeout: processor timeout.
        """
        super().__init__(pool, cache_size=cache_size, shuffle=shuffle, timeout=timeout)
        assert fnames
        self.fnames = fnames

    def iterate_unprocessed(self):
        """
        Loads unprocessed examples.

        Returns:
            Iterator of the next example.
        """
        for fname in self.fnames:
            with self.open_file(fname) as f:
                for line in f:
                    yield line

    def open_file(self, fname):
        """
        Loads lines from input files.

        Args:
            fname: text or `bz2` file to load from.
        """

        return bz2.open(fname, 'rt') if fname.endswith('.bz2') else open(fname, 'rt')



import ray
from wrangl.data import IterableDataset, Processor


class MyDataset(IterableDataset):

    strings = ['a', 'b', 'c', 'd', 'e', 'f', 'g']

    def __init__(self, pool: ray.util.ActorPool, cache_size: int = 1024, shuffle: bool = False, timeout: int = 10):
        super().__init__(self.strings, pool, cache_size, shuffle, timeout)


@ray.remote
class MyProcessor(Processor):

    def process(self, raw):
        return raw * 10


def run_no_process():
    out = []
    for s in MyDataset.strings:
        out.append(s * 10)
    return out


def run_ordered():
    pool = ray.util.ActorPool([MyProcessor.remote() for _ in range(3)])
    d = MyDataset(pool, cache_size=5, shuffle=False)
    return list(d)


def run_unordered():
    pool = ray.util.ActorPool([MyProcessor.remote() for _ in range(3)])
    d = MyDataset(pool, cache_size=5, shuffle=True)
    return list(d)


if __name__ == '__main__':
    expect = [s * 10 for s in MyDataset.strings]

    out = run_no_process()
    assert expect == out
    print(out)

    out = run_ordered()
    assert expect == out
    print(out)

    out = run_unordered()
    assert set(expect) == set(out)
    print(out)



"""
Wrangl implements two learning boilerplates:

- `wrangl.learn.supervised` for Supervised Learning, mostly using PytorchLightning.
- `wrangl.learn.rl` for Reinforcement Learning, mostly using Moolib and IMPALA.

.. include:: ../examples/learn/README.md
"""
from .supervised import SupervisedModel
__all__ = ['callbacks', 'metrics', 'supervised']



import gym
import hydra
from wrangl.learn.rl import MoolibVtrace
import atari_preprocessing


def create_env(flags):
    env = gym.make(  # Cf. https://brosa.ca/blog/ale-release-v0.7
        flags.env.name,
        obs_type="grayscale",  # "ram", "rgb", or "grayscale".
        frameskip=1,  # Action repeats. Done in wrapper b/c of noops.
        repeat_action_probability=flags.env.repeat_action_probability,  # Sticky actions.
        full_action_space=True,  # Use all actions.
        render_mode=None,  # None, "human", or "rgb_array".
    )

    # Using wrapper from seed_rl in order to do random no-ops _before_ frameskipping.
    # gym.wrappers.AtariPreprocessing doesn't play well with the -v5 versions of the game.
    env = atari_preprocessing.AtariPreprocessing(
        env,
        frame_skip=flags.env.num_action_repeats,
        terminal_on_life_loss=False,
        screen_size=84,
        max_random_noops=flags.env.noop_max,  # Max no-ops to apply at the beginning.
    )
    env = gym.wrappers.FrameStack(env, num_stack=4)
    return env


@hydra.main(config_path="conf", config_name="default")
def main(cfg):
    Model = MoolibVtrace.load_model_class(cfg.model, model_dir='model')
    Model.launch(cfg, create_env)


if __name__ == "__main__":
    main()



"""
Wrangl consists of two data types:

- `wrangl.data.dataset` are Torch Dataset wrappers that iterate over and process raw data
- `wrangl.data.processor` are Ray processors that transform each piece of the raw data

.. include:: ../examples/preprocess/README.md
"""
from .dataset import IterableDataset, SQLDataset, FileDataset
from .processor import Processor



"""
Callbacks that can be used during PytorchLightning training.
"""

import os
import glob
import json
import wandb
import torch
import tempfile
import pytorch_lightning as pl
from ..cloud import S3Client
from hydra.utils import get_original_cwd


class WandbTableCallback(pl.Callback):
    """
    Uploads sample predictions to Wandb.
    """

    def on_validation_end(self, trainer, model):
        wandb_logger = [exp for exp in trainer.logger.experiment if isinstance(exp, wandb.sdk.wandb_run.Run)]
        assert wandb_logger
        run = wandb.Api().run(path=wandb_logger[0].path)
        for artifact in run.logged_artifacts():
            artifact.delete(delete_aliases=True)
        table = wandb.Table(columns=['context', 'pred', 'gold'])
        for context, gen, gold in model.pred_samples:
            table.add_data(repr(context), repr(gen), repr(gold))
        wandb.log(dict(gen=table))


class S3Callback(pl.Callback):
    """
    Uploads experiment config, git diffs, logs, best checkpoints, and plots to S3.
    """

    def __init__(self, cfg):
        super().__init__()
        self.cfg = cfg
        self.client = S3Client(fcredentials=cfg.fcredentials)

    def on_save_checkpoint(self, trainer, model, checkpoint):
        if self.cfg.checkpoint:
            with tempfile.NamedTemporaryFile('w') as f:
                torch.save(checkpoint, f.name)
                f.flush()
                self.client.upload_file(self.cfg.project_id, self.cfg.experiment_id, 'last.ckpt', f.name, content_type='application/pytorch')

    def on_validation_end(self, trainer, model):
        self.client.upload_experiment(os.getcwd())
        if self.cfg.plot:
            self.client.plot_experiment(
                project_id=self.cfg.project_id,
                experiment_id=self.cfg.experiment_id,
                **self.cfg.plot
            )

        for fname in ['pred_samples.json']:
            self.client.upload_file(self.cfg.project_id, self.cfg.experiment_id, fname, fname, content_type='text/plain')

    def on_fit_start(self, trainer, model):
        for fname in glob.glob('git.*'):
            self.client.upload_file(self.cfg.project_id, self.cfg.experiment_id, fname, fname, content_type='text/plain')


class GitCallback(pl.Callback):
    """
    Dumps git diffs to work directory.
    """

    def on_init_end(self, trainer=None):
        import git
        repo = git.Repo(get_original_cwd(), search_parent_directories=True)
        with open('git.patch.diff', 'wt') as f:
            f.write(repo.git.diff(repo.head.commit.tree))
        with open('git.head.json', 'wt') as f:
            c = repo.head.commit
            json.dump(dict(hexsha=c.hexsha, message=c.message, date=c.committed_date, committer=c.committer.name), f, indent=2)



import minio
import pathlib
import tempfile
import os
import io
import json
import pandas as pd
from omegaconf import OmegaConf
from typing import Tuple, Union, List


class S3Client:
    """
    Instantiates a Minio Client.
    You must have a Minio-compatible S3 bucket as well as API access.
    The following environment variables are used as default, unless a credentials file is specified with the same content in a JSON file.

    - `WRANGL_S3_URL`: URL to API endpoint.
    - `WRANGL_S3_KEY`: API key.
    - `WRANGL_S3_SECRET`: API secret.
    - `WRANGL_S3_BUCKET`: bucket name.
    """

    def __init__(self, url=None, key=None, secret=None, bucket=None, fcredentials=None):
        """
        Args:
            url: S3 endpoint.
            key: S3 access key.
            secret: S3 secret key.
            bucket: S3 bucket name.
            fcredentials: alternatively specify these infomation in the form of `WRANGL_S3_{}` dictionary in a JSON credentials file.
        """
        if fcredentials:
            with open(fcredentials) as f:
                credentials = json.load(f)
        else:
            credentials = os.environ
        url = url or credentials['WRANGL_S3_URL']
        key = key or credentials['WRANGL_S3_KEY']
        secret = secret or credentials['WRANGL_S3_SECRET']
        self.bucket = bucket or credentials['WRANGL_S3_BUCKET']
        self.client = minio.Minio(endpoint=url, access_key=key, secret_key=secret)

    def upload_content(self, project_id: str, experiment_id: str, fname: str, content: str, content_type='application/json'):
        """
        Uploads raw content to a project's experiment into a file `fname`.

        The file will be accessible at `<WRANGL_S3_URL>/<WRANGL_S3_BUCKET>/<project_id>/<experiment_id>/<fname>`.

        Args:
            project_id: name of project.
            experiment_id: name of experiment.
            fname: name of file in S3 directory.
            content: content to upload.
            content_type: S3 content type.
        """
        data = io.BytesIO(content.encode('utf-8'))
        res = self.client.put_object(self.bucket, self.get_path(project_id, experiment_id, fname), data, length=len(content), content_type=content_type)
        return res

    def upload_file(self, project_id: str, experiment_id: str, fname: str, from_fname: str, content_type='application/json'):
        """
        Uploads a file to a project's experiment.

        The file will be accessible at `<WRANGL_S3_URL>/<WRANGL_S3_BUCKET>/<project_id>/<experiment_id>/<fname>`.

        Args:
            project_id: name of project.
            experiment_id: name of experiment.
            fname: name of file in S3 directory.
            from_fname: local path to file to upload.
            content_type: S3 content type.
        """
        return self.client.fput_object(self.bucket, self.get_path(project_id, experiment_id, fname), from_fname, content_type=content_type)

    def upload_experiment(self, dexp):
        """
        Uploads an experiment, including its config and logs.

        Args:
            dexp: local path to experiment directory.
        """
        dexp = pathlib.Path(dexp)
        config = OmegaConf.to_container(OmegaConf.load(dexp.joinpath('config.yaml')), resolve=True)
        project_id = config['project']

        # make sure collection exists for project
        df = []
        for flog in dexp.joinpath('logs').glob('*/*.csv'):
            df.append(pd.read_csv(flog))
        if df:
            logs = pd.concat(df).to_dict()
        else:
            logs = {}

        # make sure document does not exist
        experiment_id = config['name']

        config_response = self.upload_content(project_id=project_id, experiment_id=experiment_id, fname='config.json', content=json.dumps(config, indent=2))
        log_response = self.upload_content(project_id=project_id, experiment_id=experiment_id, fname='logs.json', content=json.dumps(logs, indent=2))
        return dict(config=config_response, log=log_response)

    @classmethod
    def get_path(cls, project_id, experiment_id, fname):
        """
        Returns the relative path of file in S3.

        Args:
            project_id: name of project.
            experiment_id: name of experiment.
            fname: name of file to look up.
        """
        return '{}/{}/{}'.format(project_id, experiment_id, fname)

    def download_content(self, project_id, experiment_id, fname):
        """
        Returns the JSON object corresponding to the uploaded file at `<WRANGL_S3_URL>/<WRANGL_S3_BUCKET>/<project_id>/<experiment_id>/<fname>`.
        """
        obj = json.loads(self.client.get_object(self.bucket, self.get_path(project_id, experiment_id, fname)).data.decode())
        return obj

    def get_experiment(self, project_id: str, experiment_id: str) -> Tuple[OmegaConf, pd.DataFrame]:
        """
        Returns the experiment's config and logs.

        Args:
            project_id: name of project.
            experiment_id: name of experiment.
        """
        config = self.download_content(project_id, experiment_id, 'config.json')
        logs = self.download_content(project_id, experiment_id, 'logs.json')
        df = pd.DataFrame(logs)
        return config, df

    def plot_experiment(self, project_id: str, experiment_id: str, x='step', y: Union[str, List[str]] = 'loss', **plot_kwargs):
        """
        Generates a line plot and uploads the result to `<WRANGL_S3_URL>/<WRANGL_S3_BUCKET>/<project_id>/<experiment_id>/<x>_vs_<y>.pdf`.
        You must have `seaborn` installed for this to work.

        Args:
            project_id: name of project.
            experiment_id: name of experiment.
            x: x axis to plot.
            y: y axis (or a list of axes) to plot.
            plot_kwargs: keyword arguments to `seaborn.lineplot`.
        """
        import seaborn as sns
        from matplotlib import pyplot as plt
        config, df = self.get_experiment(project_id=project_id, experiment_id=experiment_id)

        if not len(df):
            return

        sns.set(font_scale=1.5)

        if isinstance(y, str):
            y = [y]

        for yi in y:
            yi = yi.strip()
            fig, ax = plt.subplots(figsize=(10, 5))
            sns.lineplot(x=x, y=yi, data=df, ax=ax, **plot_kwargs)

            fname = '{}_vs_{}.pdf'.format(x, yi)
            with tempfile.TemporaryDirectory() as tempdir_path:
                tempdir = pathlib.Path(tempdir_path)
                ffig = tempdir.joinpath(fname)
                fig.savefig(ffig, bbox_inches='tight')
                self.upload_file(project_id, experiment_id, fname, ffig, content_type='application/pdf')


if __name__ == '__main__':
    mydir = pathlib.Path(os.path.abspath(__file__))
    dexp = mydir.parent.parent.joinpath('examples', 'learn', 'xor_clf', 'saves', 'wrangle-example-xor-clf', 'mymodel-default')
    client = S3Client()
    client.upload_experiment(dexp)
    config, df = client.get_experiment('wrangle-example-xor-clf', 'mymodel-default')



# Copyright (c) Facebook, Inc. and its affiliates.
# Adapted from https://github.com/facebookresearch/moolib
import copy
import dataclasses
import logging
import os
import pprint
import signal
import time
import omegaconf
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from argparse import Namespace
import wandb

import moolib
from moolib.examples.common import nest
from moolib.examples.common import vtrace
from moolib.examples.common import record
from moolib.examples import common

from pytorch_lightning.loggers import CSVLogger
from .callbacks import GitCallback, S3Callback
from .model import BaseModel


@dataclasses.dataclass
class LearnerState:
    model: torch.nn.Module
    optimizer: torch.optim.Optimizer
    scheduler: torch.optim.lr_scheduler.LambdaLR
    model_version: int = 0
    num_previous_leaders: int = 0
    train_time: float = 0
    last_checkpoint: float = 0
    last_checkpoint_history: float = 0
    global_stats: Optional[dict] = None

    def save(self):
        r = dataclasses.asdict(self)
        r["model"] = self.model.state_dict()
        r["optimizer"] = self.optimizer.state_dict()
        r["scheduler"] = self.scheduler.state_dict()
        return r

    def load(self, state):
        for k, v in state.items():
            if k not in ("model", "optimizer", "scheduler", "global_stats"):
                setattr(self, k, v)
        self.model_version = state["model_version"]
        self.model.load_state_dict(state["model"])
        self.optimizer.load_state_dict(state["optimizer"])
        self.scheduler.load_state_dict(state["scheduler"])

        for k, v in state["global_stats"].items():
            if k in self.global_stats:
                self.global_stats[k] = type(self.global_stats[k])(**v)


class MoolibVtrace(BaseModel):

    def initial_state(self, batch_size=1):
        if not self.use_lstm:
            return tuple()
        return tuple(
            torch.zeros(self.core.num_layers, batch_size, self.core.hidden_size)
            for _ in range(2)
        )

    @classmethod
    def compute_baseline_loss(cls, advantages):
        return 0.5 * torch.mean(advantages ** 2)

    @classmethod
    def compute_entropy_loss(cls, logits):
        policy = F.softmax(logits, dim=-1)
        log_policy = F.log_softmax(logits, dim=-1)
        entropy_per_timestep = torch.sum(-policy * log_policy, dim=-1)
        return -torch.mean(entropy_per_timestep)

    @classmethod
    def compute_policy_gradient_loss(cls, logits, actions, advantages):
        cross_entropy = F.nll_loss(
            F.log_softmax(torch.flatten(logits, 0, 1), dim=-1),
            target=torch.flatten(actions, 0, 1),
            reduction="none",
        )
        cross_entropy = cross_entropy.view_as(advantages)
        policy_gradient_loss_per_timestep = cross_entropy * advantages.detach()
        return torch.mean(policy_gradient_loss_per_timestep)

    @classmethod
    def create_scheduler(cls, optimizer, FLAGS):
        factor = FLAGS.unroll_length * FLAGS.virtual_batch_size / FLAGS.max_steps
        scheduler = torch.optim.lr_scheduler.LambdaLR(
            optimizer, lambda epoch: max(1 - epoch * factor, 0)
        )
        return scheduler

    @classmethod
    def compute_gradients(cls, FLAGS, data, learner_state, stats):
        model = learner_state.model

        env_outputs = data["env_outputs"]
        actor_outputs = data["actor_outputs"]
        initial_core_state = data["initial_core_state"]

        model.train()

        learner_outputs, _ = model(env_outputs, initial_core_state)

        # Use last baseline value (from the value function) to bootstrap.
        bootstrap_value = learner_outputs["baseline"][-1]

        # Move from env_outputs[t] -> action[t] to action[t] -> env_outputs[t].
        learner_outputs = nest.map(lambda t: t[:-1], learner_outputs)
        env_outputs = nest.map(lambda t: t[1:], env_outputs)
        actor_outputs = nest.map(lambda t: t[:-1], actor_outputs)

        rewards = env_outputs["reward"]
        if FLAGS.reward_clip:
            rewards = torch.clip(rewards, -FLAGS.reward_clip, FLAGS.reward_clip)

        # TODO: reward normalization ?

        discounts = (~env_outputs["done"]).float() * FLAGS.discounting

        vtrace_returns = vtrace.from_logits(
            behavior_policy_logits=actor_outputs["policy_logits"],
            target_policy_logits=learner_outputs["policy_logits"],
            actions=actor_outputs["action"],
            discounts=discounts,
            rewards=rewards,
            values=learner_outputs["baseline"],
            bootstrap_value=bootstrap_value,
        )

        entropy_loss = FLAGS.entropy_cost * cls.compute_entropy_loss(
            learner_outputs["policy_logits"]
        )
        pg_loss = cls.compute_policy_gradient_loss(
            learner_outputs["policy_logits"],
            actor_outputs["action"],
            vtrace_returns.pg_advantages,
        )
        baseline_loss = FLAGS.baseline_cost * cls.compute_baseline_loss(
            vtrace_returns.vs - learner_outputs["baseline"]
        )
        total_loss = entropy_loss + pg_loss + baseline_loss
        total_loss.backward()
        stats["env_train_steps"] += FLAGS.unroll_length * FLAGS.batch_size

    @classmethod
    def step_optimizer(cls, FLAGS, learner_state, stats):
        unclipped_grad_norm = nn.utils.clip_grad_norm_(learner_state.model.parameters(), FLAGS.grad_clip_norm)
        learner_state.optimizer.step()
        learner_state.scheduler.step()
        learner_state.model_version += 1
        stats["unclipped_grad_norm"] += unclipped_grad_norm.item()
        stats["optimizer_steps"] += 1
        stats["model_version"] += 1

    @classmethod
    def log(cls, FLAGS, logger, stats, step, is_global=False):
        stats_values = {}
        prefix = "global/" if is_global else "local/"
        for k, v in stats.items():
            stats_values[prefix + k] = v.result()
            v.reset()

        logging.info(stats_values)
        if not is_global:
            logger.log_metrics(stats_values, step=step)

        if FLAGS.wandb.enable:
            wandb.log(stats_values, step=step)

    @classmethod
    def save_checkpoint(cls, FLAGS, checkpoint_path, learner_state):
        tmp_path = "%s.tmp.%s" % (checkpoint_path, moolib.create_uid())

        logging.info("saving global stats %s", learner_state.global_stats)

        checkpoint = {
            "learner_state": learner_state.save(),
            "flags": omegaconf.OmegaConf.to_container(FLAGS),
        }

        torch.save(checkpoint, tmp_path)
        os.replace(tmp_path, checkpoint_path)

        logging.info("Checkpoint saved to %s", checkpoint_path)

    @classmethod
    def load_checkpoint(cls, checkpoint_path, learner_state):
        checkpoint = torch.load(checkpoint_path)
        learner_state.load(checkpoint["learner_state"])

    @classmethod
    def calculate_sps(cls, stats, delta, prev_steps):
        env_train_steps = stats["env_train_steps"].result()
        logging.info("calculate_sps %g steps in %g", env_train_steps - prev_steps, delta)
        stats["SPS"] += (env_train_steps - prev_steps) / delta
        return env_train_steps

    @classmethod
    def launch(cls, FLAGS, create_env, model_kwargs=None):
        logging.info("Flags:\n%s\n", pprint.pformat(dict(FLAGS)))
        logging.info("Save directory: %s", FLAGS.dsave)

        train_id = "%s/%s/%s" % (FLAGS.wandb.entity, FLAGS.wandb.project, FLAGS.wandb.name)

        logging.info("train_id: %s", train_id)

        envs = moolib.EnvPool(
            lambda: create_env(FLAGS),
            num_processes=FLAGS.num_actor_cpus,
            batch_size=FLAGS.actor_batch_size,
            num_batches=FLAGS.num_actor_batches,
        )

        if not os.path.isdir(FLAGS.localdir):
            os.makedirs(FLAGS.localdir)

        print("EnvPool started")

        device = 'cpu'
        if FLAGS.gpus > 0:
            device = 'cuda:0'
            # hack for moolib
            moolib_args = Namespace(**omegaconf.OmegaConf.to_container(FLAGS, resolve=True))
            moolib_args.device = device

        model_kwargs = model_kwargs or {}
        model = cls(FLAGS, **model_kwargs).to(device)
        optimizer = model.configure_optimizers()
        scheduler = cls.create_scheduler(optimizer, FLAGS)
        learner_state = LearnerState(model, optimizer, scheduler)

        model_numel = sum(p.numel() for p in model.parameters() if p.requires_grad)
        logging.info("Number of model parameters: %i", model_numel)

        if FLAGS.wandb.enable:
            wandb.init(
                project=str(FLAGS.wandb.project),
                config=omegaconf.OmegaConf.to_container(FLAGS),
                entity=FLAGS.wandb.entity,
                name=FLAGS.wandb.name,
            )

        env_states = [
            common.EnvBatchState(moolib_args, model) for _ in range(FLAGS.num_actor_batches)
        ]

        rpc = moolib.Rpc()
        rpc.set_name(FLAGS.name)
        rpc.connect(FLAGS.connect)

        rpc_group = moolib.Group(rpc, name=train_id)

        accumulator = moolib.Accumulator(
            group=rpc_group,
            name="model",
            parameters=model.parameters(),
            buffers=model.buffers(),
        )
        accumulator.set_virtual_batch_size(FLAGS.virtual_batch_size)

        learn_batcher = moolib.Batcher(FLAGS.batch_size, device, dim=1)

        stats = {
            "mean_episode_return": common.StatMean(),
            "mean_episode_step": common.StatMean(),
            "SPS": common.StatMean(),
            "env_act_steps": common.StatSum(),
            "env_train_steps": common.StatSum(),
            "optimizer_steps": common.StatSum(),
            "running_reward": common.StatMean(),
            "running_step": common.StatMean(),
            "steps_done": common.StatSum(),
            "episodes_done": common.StatSum(),
            "unclipped_grad_norm": common.StatMean(),
            "model_version": common.StatSum(),
            "virtual_batch_size": common.StatMean(),
            "num_gradients": common.StatMean(),
        }
        learner_state.global_stats = copy.deepcopy(stats)

        checkpoint_path = FLAGS.ckpt_path

        if FLAGS.autoresume and os.path.exists(checkpoint_path):
            logging.info("Loading checkpoint: %s" % checkpoint_path)
            cls.load_checkpoint(checkpoint_path, learner_state)
            accumulator.set_model_version(learner_state.model_version)
            logging.info("loaded stats %s", learner_state.global_stats)

        global_stats_accumulator = common.GlobalStatsAccumulator(
            rpc_group, learner_state.global_stats
        )

        terminate = False
        previous_signal_handler = {}

        def signal_handler(signum, frame):
            nonlocal terminate
            logging.info(
                "Got signal %s, quitting!",
                signal.strsignal(signum) if hasattr(signal, "strsignal") else signum,
            )
            terminate = True
            previous_handler = previous_signal_handler[signum]
            if previous_handler is not None:
                previous_signal_handler[signum] = None
                signal.signal(signum, previous_handler)

        previous_signal_handler[signal.SIGTERM] = signal.signal(
            signal.SIGTERM, signal_handler
        )
        previous_signal_handler[signal.SIGINT] = signal.signal(
            signal.SIGINT, signal_handler
        )

        if torch.backends.cudnn.is_available():
            logging.info("Optimising CuDNN kernels")
            torch.backends.cudnn.benchmark = True

        # Run.
        now = time.time()
        warm_up_time = FLAGS.warmup
        prev_env_train_steps = 0
        prev_global_env_train_steps = 0
        next_env_index = 0
        last_log = now
        last_reduce_stats = now
        is_leader = False
        is_connected = False

        dout = os.getcwd()
        logger = CSVLogger(save_dir=dout, name='logs', flush_logs_every_n_steps=FLAGS.flush_logs_every_n_steps)
        logger.log_hyperparams(FLAGS)

        if FLAGS.git.enable:
            git = GitCallback()
            git.on_init_end()

        if FLAGS.s3.enable:
            s3 = S3Callback(FLAGS.s3)
            s3.on_fit_start(trainer=None, model=None)

        while not terminate:
            prev_now = now
            now = time.time()

            steps = learner_state.global_stats["env_train_steps"].result()
            if steps >= FLAGS.max_steps:
                logging.info("Stopping training after %i steps", steps)
                break

            rpc_group.update()
            accumulator.update()
            if accumulator.wants_state():
                assert accumulator.is_leader()
                accumulator.set_state(learner_state.save())
            if accumulator.has_new_state():
                assert not accumulator.is_leader()
                learner_state.load(accumulator.state())

            was_connected = is_connected
            is_connected = accumulator.connected()
            if not is_connected:
                if was_connected:
                    logging.warning("Training interrupted!")
                # If we're not connected, sleep for a bit so we don't busy-wait
                logging.info("Your training will commence shortly.")
                time.sleep(1)
                continue

            was_leader = is_leader
            is_leader = accumulator.is_leader()
            if not was_connected:
                logging.info(
                    "Training started. Leader is %s, %d members, model version is %d"
                    % (
                        "me!" if is_leader else accumulator.get_leader(),
                        len(rpc_group.members()),
                        learner_state.model_version,
                    )
                )
                prev_global_env_train_steps = learner_state.global_stats[
                    "env_train_steps"
                ].result()

                if warm_up_time > 0:
                    logging.info("Warming up for %g seconds", warm_up_time)

            if warm_up_time > 0:
                warm_up_time -= now - prev_now

            learner_state.train_time += now - prev_now
            if now - last_reduce_stats >= 2:
                last_reduce_stats = now
                global_stats_accumulator.reduce(stats)
            if now - last_log >= FLAGS.log_every_n_steps:
                delta = now - last_log
                last_log = now

                global_stats_accumulator.reduce(stats)
                global_stats_accumulator.reset()

                prev_env_train_steps = cls.calculate_sps(stats, delta, prev_env_train_steps)
                prev_global_env_train_steps = cls.calculate_sps(
                    learner_state.global_stats, delta, prev_global_env_train_steps
                )

                steps = learner_state.global_stats["env_train_steps"].result()

                cls.log(FLAGS, logger, stats, step=steps, is_global=False)
                cls.log(FLAGS, logger, learner_state.global_stats, step=steps, is_global=True)
                if s3 is not None:
                    s3.on_validation_end(trainer=None, model=None)

                if warm_up_time > 0:
                    logging.info(
                        "Warming up up for an additional %g seconds", round(warm_up_time)
                    )

            if is_leader:
                if not was_leader:
                    leader_filename = os.path.join(
                        FLAGS.dsave, "leader-%03d" % learner_state.num_previous_leaders
                    )
                    record.symlink_path(FLAGS.localdir, leader_filename)
                    logging.info(
                        "Created symlink %s -> %s", leader_filename, FLAGS.localdir
                    )
                    learner_state.num_previous_leaders += 1
                if not was_leader and not os.path.exists(checkpoint_path):
                    logging.info("Training a new model from scratch.")
                if learner_state.train_time - learner_state.last_checkpoint >= FLAGS.val_check_interval:
                    learner_state.last_checkpoint = learner_state.train_time
                    cls.save_checkpoint(FLAGS, checkpoint_path, learner_state)

                    if s3 is not None:
                        s3.on_save_checkpoint(trainer=None, model=None, checkpoint=learner_state)

            if accumulator.has_gradients():
                gradient_stats = accumulator.get_gradient_stats()
                stats["virtual_batch_size"] += gradient_stats["batch_size"]
                stats["num_gradients"] += gradient_stats["num_gradients"]
                cls.step_optimizer(FLAGS, learner_state, stats)
                accumulator.zero_gradients()
            elif not learn_batcher.empty() and accumulator.wants_gradients():
                cls.compute_gradients(FLAGS, learn_batcher.get(), learner_state, stats)
                accumulator.reduce_gradients(FLAGS.batch_size)
            else:
                if accumulator.wants_gradients():
                    accumulator.skip_gradients()

                # Generate data.
                cur_index = next_env_index
                next_env_index = (next_env_index + 1) % FLAGS.num_actor_batches

                env_state = env_states[cur_index]
                if env_state.future is None:
                    env_state.future = envs.step(cur_index, env_state.prev_action)
                cpu_env_outputs = env_state.future.result()

                env_outputs = nest.map(
                    lambda t: t.to(device, copy=True), cpu_env_outputs
                )

                env_outputs["prev_action"] = env_state.prev_action
                prev_core_state = env_state.core_state
                model.eval()
                with torch.no_grad():
                    actor_outputs, env_state.core_state = model(
                        nest.map(lambda t: t.unsqueeze(0), env_outputs),
                        env_state.core_state,
                    )
                actor_outputs = nest.map(lambda t: t.squeeze(0), actor_outputs)
                action = actor_outputs["action"]
                env_state.update(cpu_env_outputs, action, stats)
                # envs.step invalidates cpu_env_outputs
                del cpu_env_outputs
                env_state.future = envs.step(cur_index, action)

                stats["env_act_steps"] += action.numel()

                last_data = {
                    "env_outputs": env_outputs,
                    "actor_outputs": actor_outputs,
                }
                if warm_up_time <= 0:
                    env_state.time_batcher.stack(last_data)

                if not env_state.time_batcher.empty():
                    data = env_state.time_batcher.get()
                    data["initial_core_state"] = env_state.initial_core_state
                    learn_batcher.cat(data)

                    # We need the last entry of the previous time batch
                    # to be put into the first entry of this time batch,
                    # with the initial_core_state to match
                    env_state.initial_core_state = prev_core_state
                    env_state.time_batcher.stack(last_data)
        if is_connected and is_leader:
            cls.save_checkpoint(FLAGS, checkpoint_path, learner_state)
        logging.info("Graceful exit. Bye bye!")



class Processor:
    """
    An interface object for processing examples.
    """

    def process(self, *args, **kwargs):
        """Process a single example. You must implement this."""
        raise NotImplementedError()



import ray
import time
from wrangl.data import IterableDataset, Processor
import contextlib
import tqdm
import io
import unittest


@ray.remote
class MyProcessor(Processor):

    def __init__(self):
        self.nlp = stanza.Pipeline('en')

    def process(self, raw):
        return self.nlp(raw).text


if __name__ == '__main__':
    import stanza
    tc = unittest.TestCase()

    zen = io.StringIO()
    with contextlib.redirect_stdout(zen):
        import this
    text = [zen.getvalue()] * 20

    pool = ray.util.ActorPool([MyProcessor.remote() for _ in range(3)])

    # parallel
    loader = IterableDataset(text, pool, cache_size=10, shuffle=False)
    start = time.time()
    parallel_ordered_out = list(tqdm.tqdm(loader, desc='parallel ordered', total=len(text)))
    parallel_ordered = time.time() - start

    loader = IterableDataset(text, pool, cache_size=10, shuffle=True)
    start = time.time()
    parallel_unordered_out = list(tqdm.tqdm(loader, desc='parallel unordered', total=len(text)))
    parallel_unordered = time.time() - start

    # serial
    nlp = stanza.Pipeline('en')
    start = time.time()
    serial_out = [nlp(line).text for line in tqdm.tqdm(text, desc='serial', total=len(text))]
    serial = time.time() - start

    tc.assertListEqual(
        sorted(parallel_ordered_out),
        sorted(parallel_unordered_out),
        'Parallel outputs are not equal!\nOrdered:\n{}\nUnordered\n{}'.format('\n'.join(sorted(parallel_ordered_out)), '\n'.join(sorted(parallel_unordered_out))),
    )

    tc.assertListEqual(
        parallel_ordered_out,
        serial_out,
        'Parallel output is not equal to serial!\nOrdered:\n{}\nSerial\n{}'.format('\n'.join(parallel_ordered_out), '\n'.join(serial_out)),
    )

    print('parallel ordered: {}s, parallel unordered: {}s, serial: {}s'.format(parallel_ordered, parallel_unordered, serial))
